def parse_out_file(path):
    """
    Parses an out file, which is based on ``ini`` files. They have pretty
    basic syntax. The difference from what configparser is that indentation
    is preserved in multiline values. We also support comments.

    Args:
        path (str): The path to the inifile that will be read.

    Returns:
        dict[str, dict[str, str]]: The constructed out sections=>{fields=>data}
    """
    sections = {'default': {}}
    current_section = 'default'
    current_key = None
    multiline = False
    with open(path) as outfile:
        for line in outfile:
            if line.startswith('['):
                multiline = False
                current_section = line.strip()[1:-1]
                sections[current_section] = {}
            elif line.startswith('    '):
                sections[current_section][current_key] += line[4:]
            elif line.lstrip().startswith('#') or line.lstrip().startswith(';'):
                continue
            elif line.strip():
                multiline = False
                parts = line.split(":", maxsplit=1)
                current_key = parts[0].rstrip()
                sections[current_section][current_key] = ""
                if len(parts) == 1 or not parts[1].lstrip():
                    multiline = True
                else:
                    sections[current_section][current_key] += parts[1].strip()
            elif multiline:
                sections[current_section][current_key] += "\n"
    return sections


def generate_out_file(outfile, sections):
    """
    Writes data to an output file in a structured format.

    This function takes an output file target and a dictionary of sections. Each section
    is written to the file as a header followed by its corresponding fields and
    values in key-value pair format. Line breaks in values are preserved, and values
    with line breaks are split for better readability in the file.

    Args:
        outfile: An open writable file object where the sections will be written.
        sections (dict): A mapping where each key is a section name (str), and the value
            is a dictionary containing fields (keys as strings) and their corresponding
            values. Values can contain newline characters for multiline entries.

    Returns:
        None
    """
    for section_name, section in sections.items():
        outfile.write(f"[{section_name}]\n")
        for field, value in section.items():
            outfile.write(f"{field}:")
            value = str(value)
            if '\n' in value:
                outfile.write("\n")
                for line in value.split("\n"):
                    outfile.write(f"    {line}\n")
            else:
                outfile.write(f" {value}\n")



from textwrap import dedent

from pedal.assertions.feedbacks import assert_group
from pedal.assertions.runtime import *
from pedal.core.commands import suppress
from pedal.sandbox.commands import call, start_trace, get_sandbox, evaluate, CommandBlock, run
from tests.execution_helper import Execution, ExecutionTestCase, SUCCESS_MESSAGE
from pedal.sandbox.library.drafter_library import MockDrafter
import unittest


class TestDrafter(ExecutionTestCase):

    def test_basic_mocking(self):
        with Execution(dedent("""
        from drafter import *
        from dataclasses import dataclass
        @dataclass
        class State:
            pass
        @route
        def index(state: State) -> Page:
            return Page(state, ["Hello World!"])
        start_server(State())
        """)) as e:
            suppress("algorithmic")
            pass # assert_route("index")
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)

    def test_style_mocking(self):
        with Execution(dedent("""
        from drafter import *
        from dataclasses import dataclass
        from bakery import assert_equal
        @dataclass
        class State:
            pass
        @route
        def index(state: State) -> Page:
            return Page(state, [
                "Hello World!",
                bold("This is bold"),
                Span("This is a span"),
                bold(Span("This is a bolded span")),
                
            ])
        start_server(State())
        
        assert_equal(index(State()), Page(State(), [
            "Hello World!",
            "This is made up",
            "This is bold",
            "This is a span",
            "This is a bolded span"
        ]))
        """)) as e:
            suppress("algorithmic")
        self.assertIsNone(e.student.exception)
        self.assertFeedback(e, SUCCESS_MESSAGE)


"""
Classes and functions for managing the State of a variable.
"""
from pedal.types.new_types import is_subtype


def check_trace(state):
    """
    Create a list of all the types that this given state ever took on.

    Args:
        state (:py:class:`pedal.tifa.state.State`): The state to check the trace
            history of.

    Returns:
        list[:py:class:`pedal.types.new_types.Type`]: The list of all types
            that this State ever took on.
    """
    past_types = [state.type]
    for past_state in state.trace:
        past_types.extend(check_trace(past_state))
    return past_types


class State:
    """
    A representation of a variable at a particular point in time of the program.

    Attributes:
        name (str): The name of the variable, without its scope chain
        trace (list of State): A recursive definition of previous States for
                               this State.
        type (Type): The current type of this variable.
        method (str): One of 'store', 'read', (TODO). Indicates the change that
                      occurred to this variable at this State.
        position (dict): A Position dictionary indicating where this State
                         change occurred in the source code.
        read (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been read since it was last changed. If merged from a
                    diverging path, it is possible that it was "maybe" read.
        set (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been set since it was last read. If merged from a
                    diverging path, it is possible that it was "maybe" changed.
        over (str): One of 'yes', 'no', or 'maybe'. Indicates if this variable
                    has been overwritten since it was last set. If merged from a
                    diverging path, it is possible that it was "maybe" changed.
        over_position (dict): A Position indicating where the State was
                              previously set versus when it was overwritten.
    """

    def __init__(self, name, trace, type, method, position,
                 read='maybe', set='maybe', over='maybe', over_position=None):
        self.name = name
        self.trace = trace
        self.type = type
        self.method = method
        self.position = position
        self.over_position = over_position
        self.read = read
        self.set = set
        self.over = over

    def copy(self, method, position):
        """
        Make a copy of this State, copying this state into the new State's trace
        """
        return State(self.name, [self], self.type, method, position,
                     self.read, self.set, self.over, self.over_position)


    def perfect_copy(self):
        """
        Make a perfect copy of this State, copying this state into the new State's trace
        """
        return State(self.name, self.trace, self.type, self.method, self.position,
                     self.read, self.set, self.over, self.over_position)

    def __str__(self):
        """
        Create a string representation of this State.
        """
        return "{method}(r:{read},s:{set},o:{over},{type})".format(
            method=self.method,
            read=self.read[0],
            set=self.set[0],
            over=self.over[0],
            type=self.type.__class__.__name__
        )

    def __repr__(self):
        """
        Create a string representation of this State.
        """
        return str(self)

    def was_type(self, a_type):
        """
        Retrieve all the types that this variable took on over its entire
        trace.
        """
        from pedal.types.normalize import normalize_type
        a_type = normalize_type(a_type).as_type()
        past_types = check_trace(self)
        return any(is_subtype(past_type, a_type) for past_type in past_types)


from collections import defaultdict
import hashlib

def print_history_diagram(history):
    def state_id(path_id, name, index):
        return f"{name}_{path_id}_{index}"

    def hash_style(text):
        return hashlib.md5(text.encode()).hexdigest()[:6]

    grouped = defaultdict(list)
    for path_id, fqname, state in history:
        grouped[state.name].append((path_id, fqname, state))

    print("```mermaid")
    print("flowchart TD")

    for var_name, entries in grouped.items():
        print(f"  subgraph {var_name}")
        prev_state = None
        for idx, (path_id, fqname, state) in enumerate(entries):
            sid = state_id(path_id, state.name, idx)

            # Determine highlighted lines
            highlights = {}
            if prev_state:
                if state.read != prev_state.read:
                    highlights['read'] = True
                if state.set != prev_state.set:
                    highlights['set'] = True
                if state.over != prev_state.over:
                    highlights['over'] = True
                if state.type.__class__ != prev_state.type.__class__:
                    highlights['type'] = True

            read = f"**read: {state.read}**" if highlights.get('read') else f"read: {state.read}"
            set_ = f"**set: {state.set}**" if highlights.get('set') else f"set: {state.set}"
            over = f"**over: {state.over}**" if highlights.get('over') else f"over: {state.over}"
            type_ = f"**type: {state.type.__class__.__name__}**" if highlights.get('type') else f"type: {state.type.__class__.__name__}"

            label = f"{read}<br>{set_}<br>{over}<br>{type_}"
            print(f'    {sid}["{label}"]')

            if idx > 0:
                prev_sid = state_id(entries[idx - 1][0], entries[idx - 1][2].name, idx - 1)
                print(f'    {prev_sid} -->|{state.method}| {sid}')
            prev_state = state
        print("  end")

    print("```")


from dataclasses import dataclass
from pedal.sandbox.mocked import MockModuleExposing, MethodExposer
from pedal.sandbox.mocked import MockModule
from pedal.utilities.system import IS_PYTHON_36
from functools import wraps

try:
    import drafter
except:
    drafter = None


def friendly_urls(url: str) -> str:
    if url.strip("/") == "index":
        return "/"
    if not url.startswith('/'):
        url = '/' + url
    return url


@dataclass
class GenericComponent:
    def __init__(self, name, *args, **kwargs):
        self.name = name
        self.args = args
        self.kwargs = kwargs


def _load_drafter_component(name):
    @wraps(name)
    def call_drafter_component(*args, **kwargs):
        if drafter is not None:
            component = getattr(drafter, name)
            # Component Button <class 'drafter.Button'> (MockDrafter(), 'Change the Message', <function change_message>) {}
            # raise Exception(f"Component {name} {component} {args} {kwargs}")
            return component(*args[1:], **kwargs)
        else:
            return GenericComponent(name, *args, **kwargs)
    return call_drafter_component

def _update_style(name):
    @wraps(name)
    def call_drafter_style(*args, **kwargs):
        if drafter is not None:
            return getattr(drafter, name)(*args[1:], **kwargs)
        else:
            return args[0]
    return call_drafter_style


class MockDrafter(MockModuleExposing):
    """
    Mock Drafter library that can be used to capture data from the students' program execution.
    """
    expose = MethodExposer()
    UNKNOWN_FUNCTIONS = [
        'abort', 'add_website_css', 'add_website_header',
        'deploy_site', 'get_server_setting',
        'set_website_framed', 'set_website_style', 'set_website_title',
        'set_site_information'
    ]


    def __init__(self):
        super().__init__()
        self.original_routes = []
        self.routes = {}
        self._unknown_calls = []
        self.server_started = False
        self.state = None

    def __repr__(self):
        return f"MockDrafter()"

    def __str__(self):
        return f"<MockDrafter()>"

    def add_route(self, url, func):
        self.original_routes.append((url, func))
        url = friendly_urls(url)
        self.routes[url] = func

    @expose
    def start_server(self, initial_state):
        self.server_started = True
        self.state = initial_state

    @expose
    def show_debug_information(self):
        pass

    @expose
    def hide_debug_information(self):
        pass

    @expose
    def route(self, url: str = None):
        if callable(url):
            local_url = url.__name__
            self.add_route(local_url, url)
            return url

        def make_route(func):
            local_url = url
            if url is None:
                local_url = func.__name__
            self.add_route(local_url, func)
            return func

        return make_route

COMPONENTS = ['Argument', 'Box', 'BulletedList', 'Button', 'CheckBox',
              'Div', 'Division', 'Download', 'FileUpload', 'Header',
              'HorizontalRule', 'Image', 'LineBreak', 'Link', 'LinkContent',
              'MatPlotLibPlot', 'NumberedList', 'PageContent',
              'Page',
              'Pre', 'PreformattedText', 'Row', 'SelectBox', 'Span',
              'SubmitButton',
              'Table', 'Text', 'TextArea', 'TextBox']
STYLE_FUNCTIONS = [
    'bold', 'change_background_color', 'change_border',
    'change_color', 'change_height', 'change_margin', 'change_padding', 'change_text_align',
    'change_text_decoration',
    'change_text_font', 'change_text_size', 'change_text_transform', 'change_width',
    'float_left',
    'float_right',
    'italic',
    'large_font', 'monospace',
    'small_font', 'strikethrough', 'underline',
    'update_attr', 'update_style'
]

for component in COMPONENTS:
    loaded_component = _load_drafter_component(component)
    setattr(MockDrafter, component, loaded_component)
    MockDrafter.expose.add_with_name(loaded_component, component)

for style_func in STYLE_FUNCTIONS:
    loaded_style = _update_style(style_func)
    setattr(MockDrafter, style_func, loaded_style)
    MockDrafter.expose.add_with_name(loaded_style, style_func)



import json
from pedal.core.feedback import Feedback
from pedal.core.commands import set_correct
from pedal.core.report import Report
from pedal.core.scoring import Score, combine_scores


def parse_feedback(feedback):
    """

    Args:
        feedback:

    Returns:

    """
    message = feedback.message
    title = feedback.title or feedback.label
    return feedback.correct, feedback.score, message, title, feedback.fields


class FinalFeedback:
    """
    Internal class used for organizing the feedback into one place.
    Dumpable into a simple dictionary.
    """

    DEFAULT_NO_FEEDBACK_TITLE = "No Errors"
    DEFAULT_NO_FEEDBACK_MESSAGE = "No errors reported."
    DEFAULT_NO_FEEDBACK_LABEL = "set_correct_no_errors"

    # TODO: Change all these so that FinalFeedback also logs considered feedback

    def __init__(self, correct=None, score=None, category=None, label=None, title=None,
                 message=None, data=None, hide_correctness=None,
                 suppressions=None, suppressed_labels=None, success=None,
                 scores=None, scores_feedback=None, **kwargs):
        """

        Args:
            correct:
            score:
            category:
            label:
            title:
            message:
            data:
            hide_correctness:
            suppressions:
            suppressed_labels:
            success (bool): DEPRECATED. This field is now set by `correct`. The alternative parameter name is
                kept around for backwards compatibility.
        """
        self.success = self.correct = correct if correct is not None else success
        self.score = score
        self._scores = [] if scores is None else scores
        self._scores_feedback = [] if scores_feedback is None else scores_feedback
        self.category = category
        self.label = label
        self.title = title
        self.message = message
        self.data = data
        self.hide_correctness = hide_correctness
        self.positives = []
        self.instructions = []
        self.systems = []
        self.suppressions = suppressions
        self.suppressed_labels = suppressed_labels
        self.considered = []
        self.used = []
        self._unused_extra_kwargs = kwargs

    def merge(self, feedback):
        """
        Process a single piece of feedback, and modify ourselves based on the
        data found there. This includes updating the correctness, the message,
        the title, and any other attributes.

        Also keep track of the feedback that was considered, any positive
        feedback, instructional feedback, and system messages.

        If the feedback was ultimately not meant to be used, we return None.
        Otherwise, we return the feedback object to indicate that we have
        incorporated it into our final feedback.
        """
        # print(feedback)
        self.considered.append(feedback)
        # Check if we should suppress this feedback based on its
        # category and label (and also potentially fields)
        category = feedback.category.lower()
        if category in self.suppressions:
            if True in self.suppressions[category]:
                return
            else:
                looking_for = feedback.label.lower()
                if looking_for in self.suppressions[category]:
                    list_of_fields = self.suppressions[category][looking_for]
                    # Go through each of the sets of fields
                    for fields in list_of_fields:
                        # Do any of them NOT match?
                        for field, value in fields.items():
                            if feedback.fields.get(field, None) != value:
                                break
                        else:
                            # Oh hey, a match, let's skip this guy
                            return
        # Check if this label has been suppressed specifically
        if feedback.label in self.suppressed_labels:
            list_of_fields = self.suppressed_labels[feedback.label]
            # Go through each of the sets of fields
            for fields in list_of_fields:
                # Do any of them not match?
                for field, value in fields.items():
                    if feedback['fields'].get(field, None) != value:
                        break
                else:
                    return
        # Parse the feedback for its settings
        correct, partial, message, title, data = parse_feedback(feedback)
        # If it's a system message, just add it to the systems feedback list
        if feedback and feedback.category == Feedback.CATEGORIES.SYSTEM:
            self.systems.append(feedback)
        # Resolve the score, if present, making sure to handle valence
        if not feedback.unscored and feedback.score is not None:
            # Triggered Negative leads to opposite behavior for operator
            # Also untriggered positive feedback
            invert_logic = ((feedback.valence != feedback.NEGATIVE_VALENCE) == (not feedback))
            inversion = "!" if invert_logic else ""
            self._scores.append(f"{inversion}{partial}")
            self._scores_feedback.append(feedback)
            feedback.resolved_score = Score.parse(f"{inversion}{partial}").to_percent_string()
        # If this is was not triggered but had an else message, then add it to the positives list
        if not feedback and feedback.else_message:
            self.positives.append(feedback)
            return feedback
        # If this feedback was not activated, or was muted, then don't do anything with it
        if not feedback or feedback.muted:
            return
        # If this is explicitly a positive feedback, add it to the positives list and stop
        if feedback.kind == Feedback.KINDS.COMPLIMENT:
            self.positives.append(feedback)
            return feedback
        # If this is explicitly instructional, add it to the instructions list
        if feedback.kind == Feedback.KINDS.INSTRUCTIONAL:
            self.instructions.append(feedback)
        # If this feedback was correct, then update the correctness
        # Once we have a single incorrect feedback, the whole thing is incorrect
        self.success = self.correct = correct and self.correct
        # If we don't have a message yet, then use this one
        if message is not None and self.message is None:
            self.message = message
            self.title = title
            self.category = feedback.category
            self.label = feedback.label
            self.data = data
            self.used.append(feedback)
        # All done, return the feedback out of politeness
        return feedback

    def finalize(self, max_points=None):
        """
        Finalize the feedback object, setting the title and message to
        defaults if they are not already set. Also, combine the scores
        into a single score.
        """
        # If we don't have a message yet, then use the default message
        if self.message is None:
            self.title = self.DEFAULT_NO_FEEDBACK_TITLE
            self.message = self.DEFAULT_NO_FEEDBACK_MESSAGE
        # If we have suppressed correctness, then update that flag
        self.hide_correctness = self.suppressions.get('correct', self.suppressions.get('success', False))
        # As long as we are allowed, change the default message to the "correct" message
        if (not self.hide_correctness and
                self.label == self.DEFAULT_NO_FEEDBACK_LABEL and
                self.category == Feedback.CATEGORIES.COMPLETE):
            # TODO: Promote to be its own atomic feedback function
            self.title = set_correct.title
            self.message = set_correct.message_template
            self.score = combine_scores(self._scores)
            self.success = self.correct = True
        else:
            # If they weren't correct, we need to combine the scores
            self.score = combine_scores(self._scores)
        if max_points is None:
            self.score = min(1.0, self.score)
        elif max_points != 'calculate':
            self.score = min(float(max_points), self.score)
        # Update the success/correct flags
        self.success = self.correct = bool(self.correct)
        return self

    def __str__(self) -> str:
        return "FinalFeedback({label!r}, {title!r}, {message!r})".format(label=self.label,
                                                                         title=self.title,
                                                                         message=self.message[:50])

    def __repr__(self) -> str:
        return str(self)

    def for_console(self) -> str:
        return "{label}\n{score}\n{title}\n{message}".format(label=self.label,
                                                             score=self.score,
                                                             title=self.title,
                                                             message=self.message)

    def to_json(self) -> dict:
        """

        Returns:

        """
        return {
            'correct': self.correct,
            'success': self.correct,
            'score': self.score,
            "scores": self._scores,
            'category': self.category,
            'label': self.label,
            'title': self.title,
            'message': self.message,
            'data': self.data,
            'hide_correctness': self.hide_correctness,
            'location': self.data.get('location', None) if self.data else None
        }

    def copy(self):
        return FinalFeedback(**self.to_json())

    def to_file(self, path, format='json'):
        with open(path, 'w') as f:
            if format == 'json':
                json.dump(self.to_json(), f)
            else:
                f.write(self.for_console())

def set_correct_no_errors(report: Report) -> FinalFeedback:
    return FinalFeedback(correct=True, score=0,
                         title=None, message=None,
                         category=Feedback.CATEGORIES.COMPLETE,
                         label=FinalFeedback.DEFAULT_NO_FEEDBACK_LABEL,
                         data=[], hide_correctness=False,
                         suppressions=report.suppressions,
                         suppressed_labels=report.suppressed_labels)



from pedal.types.new_types import ModuleType, FunctionType, void_function, ClassType, \
    InstanceType, bool_function, int_function, str_function, register_builtin_module, AnyType, ListType, \
    NoneType

# TODO: route, Button, ...

_VOID_FUNCTIONS = ['start_server', 'show_debug_information', 'hide_debug_information',
                   "deploy_site", "assert_equal", "set_website_framed",
                   "set_website_style", "set_website_title",
                   "add_website_header", "add_website_css", "set_site_information"]
_COMPONENT_FUNCTIONS = ['Argument', 'Box', 'BulletedList', 'Button', 'CheckBox',
              'Div', 'Division', 'Download', 'FileUpload', 'Header',
              'HorizontalRule', 'Image', 'LineBreak', 'Link',
              'MatPlotLibPlot', 'NumberedList', 'PageContent', 'SubmitButton',
              'Pre', 'PreformattedText', 'Row', 'SelectBox', 'Span',
              'Table', 'Text', 'TextArea', 'TextBox']
_IDENTITY_FUNCTIONS = ["route", 'bold',
                       'change_background_color', 'change_border', 'change_color', 'change_height',
                       'change_margin', 'change_padding', 'change_text_align',
                       'change_text_decoration', 'change_text_font', 'change_text_size',
                       'change_text_transform', 'change_width', 'float_left', 'float_right',
                       'italic', 'large_font', 'monospace', 'small_font', 'strikethrough',
                       'underline', 'update_attr', 'update_style']

def build_drafter_module():
    _PageObject = ClassType('Page', {
        'state': AnyType(),
        'content': ListType(AnyType())
    }, [])

    def _PageObject_Constructor():
        return InstanceType(_PageObject)

    _ComponentObject = ClassType('Component', {}, [])

    def _ComponentObject_Constructor():
        return InstanceType(_ComponentObject)

    _DRAFTER_FIELDS = {
        'Page': FunctionType('Page', returns=_PageObject_Constructor),
        'get_server_setting': FunctionType('get_server_setting', returns=AnyType()),
        'assert_equal': FunctionType(name='assert_equal', returns=NoneType),
    }

    _DRAFTER_FIELDS.update({
        name: void_function(name) for name in _VOID_FUNCTIONS
    })
    _DRAFTER_FIELDS.update({
        name: FunctionType(name, returns=_ComponentObject_Constructor)
        for name in _COMPONENT_FUNCTIONS
    })
    _DRAFTER_FIELDS.update({
        name: FunctionType(name, returns='identity')
        for name in _IDENTITY_FUNCTIONS
    })
    return ModuleType('drafter', _DRAFTER_FIELDS, redefines={"assert_equal"})


register_builtin_module('drafter', build_drafter_module)



from pedal.types.new_types import ModuleType, FunctionType, void_function, ClassType, \
    InstanceType, bool_function, int_function, str_function, register_builtin_module, \
    TupleType, IntType, StrType, BoolType, DictType, LiteralInt, LiteralStr, \
    FloatType, ListType

_VOID_FUNCTIONS = []
_IMAGE_FUNCTIONS = ['Image']
_IMAGE_STATIC_METHODS = [# From existing
                         'open', 'new', 'fromarray', 'frombytes', 'frombuffer',
                         # Make from scratch
                         'effect_mandelbrot', 'effect_noise', 'linear_gradient',
                         'radial_gradient',
                         # Function versions of methods
                         'alpha_composite', 'blend', 'composite', 'eval',
                         'merge', 'getchannel'
                         ]
_IMAGE_IMAGE_METHODS = ['apply_transparency', 'convert', 'copy',
                        'crop', 'effect_spread', 'filter', 'point',
                        'quantize', 'reduce', 'remap_palette', 'resize',
                        'rotate', 'transform', 'transpose',]
_IMAGE_VOID_METHODS = ['alpha_composite', 'apply_transparency',
                       'frombytes', 'paste', 'putalpha', 'putdata',
                       'putpalette', 'putpixel','save', 'seek', 'show',
                       'thumbnail', 'verify', 'load', 'close']
# TODO: draft, tobytes, tobitmap


def build_pillow_module():

    def _Image_constructor():
        return InstanceType(_Image)

    _ImagePalette = ClassType('ImagePalette', {}, [])
    _Exif = ClassType('Exif', {}, [])
    _ImagePointHandler = ClassType('ImagePointHandler', {}, [])
    _ImagePointTransform = ClassType('ImagePointTransform', {}, [])
    _ImageTransformHandler = ClassType('ImageTransformHandler', {}, [])
    _Transpose = ClassType('Transpose', {
        "FLIP_LEFT_RIGHT": LiteralInt(0),
        "FLIP_TOP_BOTTOM": LiteralInt(1),
        "ROTATE_90": LiteralInt(2),
        "ROTATE_180": LiteralInt(3),
        "ROTATE_270": LiteralInt(4),
        "TRANSPOSE": LiteralInt(5),
        "TRANSVERSE": LiteralInt(6),
    }, [])
    _Transform = ClassType("Transform", {
        "AFFINE": LiteralStr("AFFINE"),
        "EXTENT": LiteralStr("EXTENT"),
        "QUAD": LiteralStr("QUAD"),
        "MESH": LiteralStr("MESH"),
        "PERSPECTIVE": LiteralStr("PERSPECTIVE"),
    }, [])
    _Resample = ClassType("Resample", {
        "NEAREST": LiteralInt(0),
        "BOX": LiteralInt(4),
        "BILINEAR": LiteralInt(2),
        "HAMMING": LiteralInt(5),
        "BICUBIC": LiteralInt(3),
        "LANCZOS": LiteralInt(1),
    }, [])
    _Dither = ClassType("Dither", {
        "NONE": LiteralInt(0),
        "FLOYDSTEINBERG": LiteralInt(1),
        "ORDERED": LiteralInt(2),
        "RASTERIZE": LiteralInt(3),
    }, [])
    _Palette = ClassType("Palette", {
        "WEB": LiteralInt(0),
        "ADAPTIVE": LiteralInt(1),
    }, [])
    _Quantize = ClassType("Quantize", {
        "MEDIANCUT": LiteralInt(0),
        "MAXCOVERAGE": LiteralInt(1),
        "FASTOCTREE": LiteralInt(2),
        "LIBIMAGEQUANT": LiteralInt(3),
    }, [])

    _CLASS_FIELDS: dict = {
        # Fields
        'size': TupleType([IntType(), IntType()]),
        'width': IntType(),
        'height': IntType(),
        'mode': StrType(),
        'format': StrType(),
        'filename': StrType(),
        'palette': InstanceType(_ImagePalette),
        'info': DictType(),
        'is_animated': BoolType(),
        'n_frames': IntType(),
        'has_transparency_data': BoolType(),
        # Weird methods
        # TODO: How do you specify a tuple of unknown number of integers?
        'entropy': FunctionType('entropy', returns=FloatType),
        'getbands': FunctionType('getbands', returns=lambda: TupleType([StrType()])),
        'getbbox': FunctionType('getbbox', returns=lambda: TupleType([IntType(), IntType(), IntType(), IntType()])),
        'getcolors': FunctionType('getcolors', returns=lambda: ListType(element_type=TupleType([IntType(), FloatType()]))),
        'getexif': FunctionType('getexif', returns=lambda: InstanceType(_Exif)),
        'getpalette': FunctionType('getpalette', returns=lambda: ListType(element_type=IntType())),
        'getextrema': FunctionType('getextrema', returns=lambda: TupleType([FloatType(), FloatType()])),
        'getpixel': FunctionType('getpixel', returns=lambda: ListType(element_type=IntType())),
        'tell': FunctionType('tell', returns=lambda: IntType()),

    }
    _Image = ClassType('Image', _CLASS_FIELDS, [])
    _CLASS_FIELDS.update({
        name: FunctionType(name, returns=_Image_constructor)
        for name in _IMAGE_IMAGE_METHODS
    })
    _CLASS_FIELDS.update({
        name: void_function(name)
        for name in _IMAGE_VOID_METHODS
    })


    _IMAGE_MODULE_FUNCTIONS: dict = {
        'Image': _Image,
        'ImagePalette': _ImagePalette,
        'Exif': _Exif,
        'ImagePointHandler': _ImagePointHandler,
        'ImagePointTransform': _ImagePointTransform,
        'ImageTransformHandler': _ImageTransformHandler,
        'Transpose': _Transpose,
        'Transform': _Transform,
        'Resample': _Resample,
        'Dither': _Dither,
        'Palette': _Palette,
        'Quantize': _Quantize,
    }
    _IMAGE_MODULE = ModuleType('Image', fields=_IMAGE_MODULE_FUNCTIONS)
    _IMAGE_MODULE_FUNCTIONS.update({
        name: FunctionType(name, returns=_Image_constructor)
        for name in _IMAGE_STATIC_METHODS
    })

    return ModuleType('PIL', {}, {'Image': _IMAGE_MODULE})


register_builtin_module('PIL', build_pillow_module)



import os
import sys
import io
from contextlib import redirect_stdout
import unittest

from pedal import MAIN_REPORT, call
from pedal.utilities.system import IS_AT_LEAST_PYTHON_311

pedal_library = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, pedal_library)

here = "" if os.path.basename(os.getcwd()) == "tests" else "tests/"

from tests.execution_helper import ExecutionTestCase
from pedal.environments.standard import setup_environment
from pedal.source import next_section, verify_section, check_section_exists, set_source_file, separate_into_sections
from pedal.core.commands import compliment, clear_report, contextualize_report, log, explain
from pedal.assertions import assert_equal, prevent_ast


class TestEnvironmentStandard(ExecutionTestCase):

    def checkFeedback(self, actual, expected):
        self.assertEqual(expected.strip(), actual.strip())

    def test_correct(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="""1+2""")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
Title: Complete
Label: set_correct_no_errors
Score: 1
Message: Great work!
''')

    def test_empty_source(self):
        with io.StringIO() as f, redirect_stdout(f):
            terminal = setup_environment(main_code="")
            terminal.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
Title: No Source Code
Label: blank_source
Score: 0
Message: Source code file is blank.
''')

    def test_compliments(self):
        with io.StringIO() as f, redirect_stdout(f):
            standard = setup_environment(main_code="1+2")
            compliment("Good job!")
            standard.resolve()
            clear_report()
            self.checkFeedback(f.getvalue(), '''
Title: Complete
Label: set_correct_no_errors
Score: 1
Message: Great work!
Compliments:
  - Good job!
''')